// ============================================================================
// Objektorientierte Programmierung - Hausaufgabe 9 - Feuerwerk
// Name:
// ============================================================================

// Bedienung des Programms:
// -----------------------
// 	Wenn man die Taste 'q' oder 'Q' drueckt, wird das Programm beendet.
// 	Wenn man ' ' (Leerzeichen) drueckt, wird die Animation pausiert.
// 	Wenn man 'Enter'/'Return' drueckt, startet die Animation neu.

// Fuer graphische Darstellung:
import java.awt.Color;

import java.awt.*;
import java.awt.event.*;

// ============================================================================
// Klasse fuer animierte graphische Dartstellung (erzeugt Fenster):
// ============================================================================

class Fireworks extends Frame implements WindowListener, KeyListener, Runnable{

	// --------------------------------------------------------------------
	// Konstanten (waehlbare Konfigurations-Parameter):
	// --------------------------------------------------------------------

	// Fenster-Groesse:
	public static final int X_SIZE = 500;
	public static final int Y_SIZE = 500;

	// Hintergrund-Farbe:
	private static final Color BACKGROUND = Color.BLACK;

	// Hunderstel Sekunden zwischen zwei Bildern:
	private static final int TIME_STEP = 5; // d.h. 20 Bilder / Sekunde

	// Verwendung eines doppelten Puffers, um Flackern zu reduzieren:
	private static final boolean DOUBLE_BUFFER = true;

	// Zeichnen von Kreisen fuer Leuchtkugeln (sonst Quadrate)
	private static final boolean DRAW_CIRCLE = true;
		// Kreise brauchen etwas mehr Rechenzeit

	// --------------------------------------------------------------------
	// Attribute der Klasse:
	// --------------------------------------------------------------------

	// Attribut fuer zu zeichnende Leuchtkugeln:
	private Star[] stars;

	// Attribut fuer aktuelle Zeit:
	private int currTime;

	// Zeitpunkt des letzten Bildes (muss geloescht werden):
	private int lastTime;

	// Dauer der Animation (maximale Zeit):
	private int duration;

	// Ist die Animation gerade angehalten ("Pause")?
	private boolean isPaused;

	// Attribut fuer Thread zur parallelen Ausfuehrung der Animation:
	private Thread thread;

	// Eindeutige Nummer fuer Thread, der repaint-Ereignisse erzeugen soll:
	private int threadId;

	// Offscreen Image fuer "Double Buffering" Methode:
	private Image bufImage;

	// Graphic Context fuer "Offscreen Image" (bufImage):
	private Graphics bufGraphics;

	// Groesse des "Offscreen Image" (bufImage):
	private int bufWidth;
	private int bufHeight;

	// --------------------------------------------------------------------
	// Konstruktor:
	// --------------------------------------------------------------------

	public Fireworks(Star[] stars, int duration) {
		// Konstruktor der Oberklasse aufrufen (Titelzeile Fenster):
		super("Feuerwerk");

		// Testausgabe:
		for(int i = 0; i < stars.length; i++) {
			System.out.println(stars[i]);
		}

		// Parameterwerte speichern:
		this.stars = stars;
		this.duration = duration;

		// Weitere Initialisierungen (eigentlich ueberfluessig):
		this.currTime = 0;
		this.lastTime = 0;
		this.isPaused = false;
		this.thread = null;
		this.threadId = 0;
		this.bufImage = null;
		this.bufWidth = 0;
		this.bufHeight = 0;
		this.bufGraphics = null;

		// Fensterereignisse bitte hierher zustellen:
		addWindowListener(this);

		// Ereignisse fuer Tastendruecke bitte auch:
		addKeyListener(this);

		// Fensterdaten setzen und Fenster sichtbar machen:
		setBackground(Color.BLACK);
		setSize(X_SIZE, Y_SIZE);
		setVisible(true);
	}

	// --------------------------------------------------------------------
	// Ereignis-Behandlung (Window-Listener fuer Fenster-Ereignisse):
	// --------------------------------------------------------------------

	// Verschiedene Fenster_Ereignisse (ohne Reaktion):
	public void windowActivated(WindowEvent e) {}
	public void windowDeactivated(WindowEvent e) {}
	public void windowIconified(WindowEvent e) {}
	public void windowDeiconified(WindowEvent e) {}
	public void windowOpened(WindowEvent e) {}

	// Benutzer will Fenster schliessen:
	public void windowClosing(WindowEvent e) {
		e.getWindow().setVisible(false);
		e.getWindow().dispose();
	}
	// Wenn Fenster geloescht ist, endet das Programm:
	public void windowClosed(WindowEvent e) {
		System.exit(0);
	}

	// --------------------------------------------------------------------
	// Ereignis-Behandlung (Key-Listener fuer gedrueckte Tasten):
	// --------------------------------------------------------------------

	public void keyPressed(KeyEvent e) {}
	public void keyReleased(KeyEvent e) {}

	public void keyTyped(KeyEvent e) {
		char key = e.getKeyChar();

		// Pause-Funktion:
		if(key == ' ') {
			if(this.isPaused)
				startAnimation();
			else
				stopAnimation();
			this.isPaused = ! this.isPaused;
		}

		// Neubeginn der Simulation:
		else if(key == '\n') {
			stopAnimation();
			this.currTime = 0;
			this.isPaused = false;
			startAnimation();
		}

		// Quit:
		else if(key == 'q' || key == 'Q') {
			this.setVisible(false);
			this.dispose();
			System.exit(0);
		}
	}

	// --------------------------------------------------------------------
	// Paralleler Thread fuer Animation:
	// --------------------------------------------------------------------

	// Thread fuer Animation erzeugen (erzeugt Ereignisse zum Neu-Zeichnen)
	public void startAnimation() {
		thread = new Thread(this);
		thread.start();
	}

	// Dies ist, was der parallele Thread macht:
	public void run() {
		int id = this.threadId;
		while(this.currTime < this.duration && id == this.threadId) {
			repaint();
			try {
				Thread.sleep(TIME_STEP * 10);
			} catch(InterruptedException e) {
				// Wir machen einfach normal weiter
			}
			this.currTime += TIME_STEP;
		}
	}

	// Zum Anhalten der Simulation diese Methode aufrufen:
	public void stopAnimation() {
		this.threadId++;
		// run() bemerkt dann, dass eigene id nicht mehr aktuell
	}

	// --------------------------------------------------------------------
	// Behandlung ungueltiger Koordinaten etc.:
	// --------------------------------------------------------------------

	// Fehlermeldung und Schluss:
	private static void error(int i, int t, String msg, int val) {
		System.err.println("Fehler (Stern " + i + ", t=" + t + "): " +
					msg + " (" + val + ")");
		System.exit(1);
	}

	// --------------------------------------------------------------------
	// Eigentliche graphische Ausgabe:
	// --------------------------------------------------------------------

	// Leuchtstern zeichnen:
	private void drawStar(Graphics g, int x, int y, int size,
							int i, int t) {
		// Koordinaten pruefen:
		if(x < 0 || x > X_SIZE)
			error(i, t, "x-Wert ungueltig", x);
		if(y < 0 || y > Y_SIZE)
			error(i, t, "y-Wert ungueltig", y);
		if(size < 0 || size > X_SIZE/2 || size > Y_SIZE/2)
			error(i, t, "Groesse ungueltig", size);

		if(DRAW_CIRCLE)
			g.fillOval(x-size, y-size, 2*size, 2*size);
		else
			g.fillRect(x-size, y-size, 2*size, 2*size);
	}

	// Loeschen des Bildes verhindern, um Flackern zu reduzieren:
	@Override public void update(Graphics g) {
		if(!DOUBLE_BUFFER)
			paint(g);
		else {
			// Ggf. Offscreen-Puffer initialisieren:
			if(bufImage == null) {
				bufWidth = this.getSize().width;
				bufHeight = this.getSize().height;
				bufImage = createImage(bufWidth, bufHeight);
				bufGraphics = bufImage.getGraphics();
			}

			// Hintergrund loeschen:
			bufGraphics.setColor(BACKGROUND);
			bufGraphics.fillRect(0, 0, bufWidth, bufHeight);

			// Auf den Offscreen-Puffer zeichnen:
			bufGraphics.setColor(getForeground());
			paint(bufGraphics);

			// Offscreen Puffer in sichtbaren Bildschirm kopieren:
			g.drawImage(bufImage, 0, 0, this);
		}
	}

	// Hier wird gezeichnet:
	@Override public void paint(Graphics g) {


		// Alle Leuchtsterne zum vorigen Zeitpunkt loeschen:
		if(!DOUBLE_BUFFER) {
			// Zeitpunkt des vorigen Bildes:
			int t = this.lastTime;

			// Farbe des Zeichenstiftes setzen:
			g.setColor(BACKGROUND);

			// Jetzt alle Leuchtkugeln ausschwaerzen:
			for(int i = 0; i < this.stars.length; i++) {

				// War der Leuchtstern ueberhaupt sichtbar?
				if(stars[i].visible(t)) {

					// Daten des i-ten Leuchtsterns holen:
					int x = stars[i].posX(t);
					int y = stars[i].posY(t);
					int size = stars[i].size(t);

					// Zeichnen (in Hintergrund-Farbe):
					drawStar(g, x, y, size, i, t);
				}
			}
		}

		// Alle Leuchtsterne zum aktuellen Zeitpunkt zeichnen:
		for(int i = 0; i < this.stars.length; i++) {

			// Aktueller Zeitpunkt:
			int t = this.currTime;

			// Ist der i-te Leuchtstern ueberhaupt sichtbar?
			if(stars[i].visible(t)) {

				// Daten des i-ten Leuchtsterns holen:
				int x = stars[i].posX(t);
				int y = stars[i].posY(t);
				int size = stars[i].size(t);
				Color color = stars[i].color(t);

				// Farbe des Zeichenstifts setzen:
				g.setColor(color);

				// Zeichnen:
				drawStar(g, x, y, size, i, t);
			}
		}

		// Letzten dargestellten Zeitpunkt aktualisieren:
		this.lastTime = this.currTime;
	}

	// --------------------------------------------------------------------
	// Hauptprogramm:
	// --------------------------------------------------------------------

	public static void main(String[] args) {

		// Feuerwerksdaten holen:
		Star[] s = FireworksShow.stars();
		int d = FireworksShow.duration();

		// Fenster erzeugen und Animation anzeigen:
		Fireworks f = new Fireworks(s, d);
		f.startAnimation();
	}
}

// ============================================================================
// Abstrakte Oberklasse fuer bewegte Leuchtkugeln:
// ============================================================================

abstract class Star {
	// Die Zeitpunkte sind in Hunderstel Sekunden.
	// Wenn 20 Bilder pro Sekunde erzeugt werden,
	// wird nur jeder 5-te Wert angefragt.

	// Ist das Objekt sichtbar zum Zeitpunkt t?
	abstract boolean visible(int t);

	// Wenn das Objekt zum Zeitpunkt t nicht sichtbar
	// ist, spielen die Ergebnisse der folgenden
	// Methoden keine Rolle.

	// x-Position (Bereich 0 .. 500) zum Zeitpunkt t:
	abstract int posX(int t);

	// y-Position (Bereich 0 .. 500) zum Zeitpunkt t:
	abstract int posY(int t);

	// Farbe des Objektes zum Zeitpunkt t:
	abstract Color color(int t);

	// Radius der Leuchtkugel zum Zeitpunkt t (Pixel):
	abstract int size(int t);
}

// ============================================================================
// Hier Ihre Unterklasse(n) von Star (Leuchtkugeln):
// ============================================================================

	// Mindestens die Unterklasse SimpleStar fuer
	// Leuchtkugeln mit geradlinier Flugbahn und konstanter Geschwindigkeit
	//
	// Der Konstruktor soll 7 Parameter haben:
	// - Zeitintervall, in dem die Leuchtkugel sichtbar ist [t1, t2]
	// - Koordinaten der Leuchtkugel zum Zeitpunkt t1: (x1,y1)
	// - Koordinaten der Leuchtkugel zum Zeitpunkt t2: (x2, y2)
	// - Farbe der Leuchtkugel

class SimpleStar extends Star {

	// Attribute:
	private int startTime;
	private int stopTime;
	private int startX;
	private int startY;
	private int stopX;
	private int stopY;
	private Color color;

	// Konstruktor:
	public SimpleStar(int startTime, int stopTime,
			int startX, int startY,
			int stopX, int stopY,
			Color color) {
		this.startTime = startTime;
		this.stopTime  = stopTime;
		this.startX = startX;
		this.startY = startY;
		this.stopX = stopX;
		this.stopY = stopY;
		this.color = color;
	}

	// Ist das Objekt sichtbar zum Zeitpunkt t?
	boolean visible(int t) {
		return (t >= startTime && t < stopTime);
	}

	// x-Position (Bereich 0 .. 500) zum Zeitpunkt t:
	int posX(int t) {
		if(stopTime > startTime)
			return startX +
				(t - startTime) * (stopX - startX) /
					(stopTime - startTime);
		else
			return 0;
	}

	// y-Position (Bereich 0 .. 500) zum Zeitpunkt t:
	int posY(int t) {
		if(stopTime > startTime)
			return startY +
				(t - startTime) * (stopY - startY) /
					(stopTime - startTime);
		else
			return 0;
	}

	// Farbe des Objektes zum Zeitpunkt t:
	Color color(int t) {
		return color;
	}

	// Radius der Leuchtkugel zum Zeitpunkt t (Pixel):
	int size(int t) {
		return 4;
	}

	// Fuer Testausgaben:
	@Override public String toString() {
		return String.format (
		"SimpleStar: [%3d, %3d] (%3d, %3d) -> (%3d, %3d) " +
				"Color: R=%d,G=%d,B=%d",
				startTime, stopTime,
				startX, startY,
				stopX, stopY,
				color.getRed(), color.getGreen(),
				color.getBlue());
	}
}

// ============================================================================
// Klasse mit den Daten des Feuerwerks:
// ============================================================================

class FireworksShow {

	// Diese Methode liefert das Array mit den Leuchtkugel-Objekten:
	static Star[] stars() {
		// Parameter der Feuerwerksbombe:
		int startHeight = 0;
		int burstHeight = 350;
		int startX = 250;
		Color cometColor = Color.RED;
		Color starColor = Color.GREEN;
		int cometSize = 4;	// Wird von SimpleStar ignoriert
		int starSize = 2;	// Wird von SimpleStar ignoriert
		int numStars = 8;
		int spreadSize = 100;
		int startTime = 0;
		int burstTime = 300;
		int spreadTime = 100;

		// Abgeleitete Werte:
		int startY = Fireworks.Y_SIZE - startHeight;
		int burstY = Fireworks.Y_SIZE - burstHeight;
		int burstX = startX;
		int stopTime = burstTime + spreadTime;

		// Array, in das die Leuchtkugeln gespeichert werden:
		Star[] s = new Star[numStars+1];

		// Roter Leuchtstern als Aufstiegseffekt:
		s[0] = new SimpleStar(startTime, burstTime,
					startX, startY,
					burstX, burstY,
					cometColor);

		// Platzt kreisfoermig in 8 gruene Leuchtsterne:
		for(int i = 0; i < numStars; i++) {
			double angle = Math.toRadians(360*i/(double)numStars);
			int stopX = burstX +
					(int) Math.round (spreadSize *
						Math.cos(angle));
			int stopY = burstY +
					(int) Math.round (spreadSize *
							Math.sin(angle));
			s[i+1] = new SimpleStar(burstTime, stopTime,
					burstX, burstY,
					stopX, stopY,
					starColor);
		}
		return s;
	}

	// Gewuenschte Gesamt-Dauer der Simulation:
	static int duration() {
		return 501;
	}
}

